////////////////////////////////////////////////////////
////////////////////////////////////////////////////////
////    main.c 
////////////////////////////////////////////////////////
////////////////////////////////////////////////////////


#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <avr/io.h>
#include <avr/pgmspace.h>

/* Scheduler include files. */
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"

/* Pololu include files. */
#include <pololu/orangutan.h>

/* Clock include file. */
#include "retrograde.h"

/* i2c Interface include file. */
#include "i2cmaster.h"

/* Servo PWM Timer include file */
#include "ServoPWMTimer.h"


/* Main program loop */

int main(void) 
{
    
    xADCQueue = xQueueCreate( 2, sizeof( xADCArray ) );  // queue for ADC Sensor values
    
    xI2CRTCQueue = xQueueCreate( 2, sizeof( xRTCArray) );  // queue for i2c RTC values
    
    xADCSemaphore = xSemaphoreCreateMutex(); // mutex semaphore for ADC    
    xI2CSemaphore = xSemaphoreCreateMutex(); // mutex semaphore for i2c bus
    xLCDSemaphore = xSemaphoreCreateMutex(); // mutex semaphore for LCD
    xRetrogradeSemaphore = xSemaphoreCreateMutex(); // mutex semaphore for Retrograde Analogue Hands
    
    set_digital_output(IO_B5, LOW); // initialise Green IO_C4 LED  (Arduino IO_B5)

    set_analog_mode(MODE_10_BIT);    // 10-bit analog-to-digital conversions
    start_analog_conversion(0);     // start initial ADC conversion on PC0
    
    i2c_init();         // init I2C interface, need to do this once only.
    
    /*  To adjust the date/time uncomment this function, and set a time in the near future in function below.
        Load the program. 4 seconds before the time set press the reset button.
        Then comment out this line, and reload the program.
    */
//    setDateDs1307();    // set the date & time. Set the correct values, and it will run once.

    clear();			// clear the LCD, move cursor to start of top line   
    
    start_PWM_hardware();  // start the PWM TimerX hardware depending on the Timer #define in FreeRTOSConfig.h
    
    xTaskCreate(
		TaskBlinkGreenLED
		,  (const signed portCHAR *)"GreenLED"
		,  128
		,  NULL
		,  3
		,  NULL );

    xTaskCreate(
        TaskReadADCSensors
        ,  (const signed portCHAR *)"ReadADCSensors"
        ,  128
        ,  NULL
        ,  1
        ,  NULL );

    xTaskCreate(
        TaskReadI2CRTC
        ,  (const signed portCHAR *)"ReadI2CRTC"
        ,  128
        ,  NULL
        ,  1
        ,  NULL );      

    xTaskCreate(
		TaskWriteAnalogLCD
		,  (const signed portCHAR *)"WriteAnalogLCD"
		,  128
		,  NULL
		,  2
		,  NULL );

   xTaskCreate(
		TaskWriteRTCLCD
		,  (const signed portCHAR *)"WriteRTCLCD"
		,  128
		,  NULL
		,  2
		,  NULL );

   xTaskCreate(
		TaskWriteRTCRetrograde
		,  (const signed portCHAR *)"WriteRTCRetrograde"
		,  128
		,  NULL
		,  2
		,  NULL );

    vTaskStartScheduler();

    //and finally, to eliminate compiler warnings....
    return(0);
}



/*-----------------------------------------------------------*/


static void TaskBlinkGreenLED(void *pvParameters) // Main Green LED Flash
{
    (void) pvParameters;;
    portTickType xLastWakeTime;
	/* The xLastWakeTime variable needs to be initialized with the current tick
	count.  Note that this is the only time we access this variable.  From this
	point on xLastWakeTime is managed automatically by the vTaskDelayUntil()
	API function. */
	xLastWakeTime = xTaskGetTickCount();
    
    while(1)
    {
    set_digital_output( IO_B5, 1);               // main (green IO_C4) LED on   (Arduino IO_B5)
	vTaskDelayUntil( &xLastWakeTime, ( 100 / portTICK_RATE_MS ) );

	set_digital_output( IO_B5, 0);               // main (green IO_C4) LED off  (Arduino IO_B5)
	vTaskDelayUntil( &xLastWakeTime, ( 300 / portTICK_RATE_MS ) );
    }
}

/*-----------------------------------------------------------*/


static void TaskReadADCSensors(void *pvParameters) // Read ADC Sensor for Thermal LM335z
{
    (void) pvParameters;;
    portTickType xLastWakeTime;
	/* The xLastWakeTime variable needs to be initialized with the current tick
	count.  Note that this is the only time we access this variable.  From this
	point on xLastWakeTime is managed automatically by the vTaskDelayUntil()
	API function. */
	xLastWakeTime = xTaskGetTickCount();

	
	/*

     * Variables for the analogue conversion on ADC Sensors
     */
     
    const char samples = 50;        // determines the number of samples taken for averaging
    unsigned int sum;               // holds the summated samples
    xADCArray values;                // holds the return values to put on queue
    unsigned char i;
	
    while(1)
    {
        if( xADCSemaphore != NULL )
        {
            // See if we can obtain the semaphore.  If the semaphore is not available
            // wait 10 ticks to see if it becomes free.	
            if( xSemaphoreTake( xADCSemaphore, ( portTickType ) 10 ) == pdTRUE )
            {
            // We were able to obtain the semaphore and can now access the shared resource.
            // We want to have the ADC for us alone, as it takes some time to sample,
            // so we don't want it getting stolen during the middle of a conversion.

                i = samples;                     // clean up for next conversion
                sum = 0;
                                    
                do
                {
                    start_analog_conversion(0);   // start next conversion 
                    while( analog_is_converting() )
                    {  
                        taskYIELD();     // yield until conversion ready
                    }
	                sum += analog_conversion_result();	// sum the results
	                ADCSRA |= 1 << ADSC;	// start the next conversion on current channel
                } while (--i);
                                              
                values.adc0 = (unsigned int) (sum / (int) samples);    // sample average of ADC result
                                                 
                             
                xSemaphoreGive( xADCSemaphore );
                       
                if( xADCQueue != 0 )
                {
                    // Send the structure.  Wait for 10 ticks for space to become
                    // available if necessary.
                    xQueueSendToBack( xADCQueue, &values, ( portTickType ) 10 );            
                }
       
                vTaskDelayUntil( &xLastWakeTime, ( 200 / portTICK_RATE_MS ) );   
            }
        }
    }
}


    
static void TaskReadI2CRTC(void *pvParameters) // Read i2c Bus for Real Time Clock DS1307
{
    (void) pvParameters;;
    
    portTickType xLastWakeTime;
	/* The xLastWakeTime variable needs to be initialized with the current tick
	count.  Note that this is the only time we access this variable.  From this
	point on xLastWakeTime is managed automatically by the vTaskDelayUntil()
	API function. */
	xLastWakeTime = xTaskGetTickCount();
	
	xRTCArray xReadings;     // Holds return values from the RTC DS1307

    unsigned char ret = 0;      // temporary variable
        
    while(1)
    {
        if( xI2CSemaphore != NULL )
        {
            // See if we can obtain the semaphore.  If the semaphore is not available
            // wait 10 ticks to see if it becomes free.	
            if( xSemaphoreTake( xI2CSemaphore, ( portTickType ) 10 ) == pdTRUE )
            {             
                // We were able to obtain the semaphore and can now access the
                // shared resource.
                        
                /*  Reading from the Slave
                1. Send a start sequence
                2. Send 0xD0 ( I2C address of the DS1307 with the R/W bit low (even address)
                3. Send 0x00 (Internal address of the bearing register)
                4. Send a start sequence again (repeated start)
                5. Send 0xD1 ( I2C address of the DS1307 with the R/W bit high (odd address)
                6. Read data byte from DS1307
                7. Repeat, reading the next data byte from DS1307                
                8. Send the stop sequence.
                */    
                

                ret = i2c_start(DS1307+I2C_WRITE);       // set device address and write mode              
                  
                if ( ret )
                {
                    /* failed to issue start condition, possibly no device found */
                    i2c_stop();
                    set_digital_output( IO_D3, 1);               // red LED on to show error */
                    
                } else {
                    /* issuing start condition ok, device accessible */
                    i2c_write(0x00);                       // write address = 0 (Seconds)
            
                    i2c_rep_start(DS1307+I2C_READ);              // set device address and read mode

                    xReadings.Seconds =  bcdToDec( i2c_readAck() & 0x7f );  // read one byte from address 0
                    xReadings.Minutes =  bcdToDec( i2c_readAck() );         // read one byte from address 1
                    xReadings.Hours   =  bcdToDec( i2c_readAck() & 0x3f );  // read one byte from address 2
                    xReadings.Day     =  bcdToDec( i2c_readAck() );         // read one byte from address 3
                    xReadings.Date    =  bcdToDec( i2c_readAck() );         // read one byte from address 4
                    xReadings.Month   =  bcdToDec( i2c_readAck() );         // read one byte from address 5
                    xReadings.Year    =  bcdToDec( i2c_readAck() );         // read one byte from address 6
                    xReadings.Control =  bcdToDec( i2c_readNak() );         // read one last byte from address 7
                    i2c_stop();                                             // set stop condition = release bus
                }
               
                // We have finished accessing the shared resource.
                // Release the i2c semaphore.
                xSemaphoreGive( xI2CSemaphore );
            }
        }
        
        if( xI2CRTCQueue != 0 )
        {         
            // Send the Tranport structure.  Wait for 10 ticks for space to become
            // available if necessary.           
            xQueueSendToBack( xI2CRTCQueue, &xReadings, ( portTickType ) 10 );  
        }                  

        vTaskDelayUntil( &xLastWakeTime, ( 200 / portTICK_RATE_MS ) );
    }
}

/*-----------------------------------------------------------*/
               

static void TaskWriteAnalogLCD(void *pvParameters) // Write to LCD
{
    (void) pvParameters;;
    
    xADCArray result;  // variable to hold the average ADC value off the queue
    
   /*
   For the LM335Z we want to calculate the resistance R1 required to ensure that we have 500 uA minimum at the maximum
   temperature we intend to measure.
   Assume maximum 60C this is 273K + 60 = 333K or will be measured at 3330mV

   If Vcc is 4.9V (USB) then the R = V/I calcuation gives (4.9 - 3.3)/ 0.0005 = 3200 Ohm
   
   This leads to using a 3200 Ohm resistor, or there abouts being 3300 Ohm.
   
   Testing gives us 0.58mA with the resistor actual at 3250 Ohm.
   
   Analogue testing gives with this set up: 2.952V at 20C actual... or 22C indicated
   
   Lets see what the Arduino ADC gives us...
   
   */ 
    
    
    while(1)
    {
        if( xADCQueue != 0 )
        {
        /* Block on the queue to wait for data to arrive. */        
            if( xLCDSemaphore != NULL )
            {
                // See if we can obtain the semaphore.  If the semaphore is not available
                // wait 10 ticks to see if it becomes free.	
                if( xSemaphoreTake( xLCDSemaphore, ( portTickType ) 10 ) == pdTRUE )
                {
                    // We were able to obtain the semaphore and can now access the
                    // shared resource.

                    if( xQueueReceive( xADCQueue, &result, portMAX_DELAY) == pdPASS )
                    {
                        lcd_goto_xy(0, 3);          // go to the first character of the forth LCD line
                        print_long( result.adc0  );            // print the averaged ADC value
                        print(" ADC  ");			// added spaces are to overwrite left over chars
                        lcd_goto_xy(9, 3);			// LCD cursor to tenth character of the forth LCD line
                        print_long((portLONG)( (portFLOAT)result.adc0 * 5000.0 / 1023.0) ); // print mV converted
                        print("mV  ");			// added spaces are to overwrite left over chars
                        lcd_goto_xy(16, 3);			// LCD cursor to seventeenth character of the forth LCD line
                        print_long((portLONG)(( (portFLOAT)result.adc0 * 500.0 / 1023.0)- 273.15)); // print mV converted
                        print("C ");			// added spaces are to overwrite left over chars                
                    }           
                }
                xSemaphoreGive( xLCDSemaphore );
                taskYIELD();                      
            }
        }
    }
}
    

    
static void TaskWriteRTCLCD(void *pvParameters) // Write RTC to LCD
{
    (void) pvParameters;;

    xRTCArray xValues;     // Holds result values from the Real Time Clock DS1307.
    
    while(1)
    {
        /* Block on the queue to wait for data to arrive. */              
        if( xI2CRTCQueue != 0 )
        {
            if( xLCDSemaphore != NULL )
            {
                // See if we can obtain the semaphore.  If the semaphore is not available
                // wait 10 ticks to see if it becomes free.	
                if( xSemaphoreTake( xLCDSemaphore, ( portTickType ) 10 ) == pdTRUE )
                {
                    // We were able to obtain the semaphore and can now access the
                    // shared resource.

                    if( xQueueReceive( xI2CRTCQueue, &xValues, portMAX_DELAY) == pdPASS )
                    {
                        
                        lcd_goto_xy(0, 0);              // go to the first character of the first LCD line
                        print("20");
                        print_long( xValues.Year );
                        print("-");
                        print_long( xValues.Month );
                        print("-");
                        print_long( xValues.Date );
                        print("  ");			         // added spaces are to overwrite left over chars
                        
                        lcd_goto_xy(12, 0);             // go to the thirteenth character of the first LCD line
                        if(xValues.Hours < 10) { print(" "); }
                        print_long( xValues.Hours);
                        if(xValues.Minutes < 10) { print(":0"); }
                        else{ print(":"); }
                        print_long( xValues.Minutes );
                        if(xValues.Seconds < 10) { print(":0"); }
                        else{ print(":"); }
                        print_long( xValues.Seconds );  
                        
                        lcd_goto_xy(5, 1);              // go to the sixth character of the second LCD line
                        switch( xValues.Day )
                        {
                            case 1:
                                print("Sun");
                                break;
                            case 2:
                                print("Mon");
                                break;
                            case 3:
                                print("Tues");
                                break;
                            case 4:
                                print("Wednes");
                                break;
                            case 5:
                                print("Thurs");
                                break;
                            case 6:
                                print("Fri");
                                break;
                            case 7:
                                print("Satur");
                                break;
                        }
                        print("day    ");                 
                    
                    }           
                }
                xSemaphoreGive( xLCDSemaphore );
                
                if(xValues.Seconds == 0)
                {   
                    xSemaphoreGive( xRetrogradeSemaphore );
                }
                taskYIELD();                      
            }
        }
    }
}


static void TaskWriteRTCRetrograde(void *pvParameters) // Write RTC to Retrograde Hands
{
    (void) pvParameters;;

    xRTCArray xValues;     // Holds result values from the Real Time Clock DS1307.

    unsigned int servoHours_us;    
    unsigned int servoMinutes_us;

    while(1)
    {
        /* Block on the queue to wait for data to arrive. */              
        if( xI2CRTCQueue != 0 )
        {
            if( xRetrogradeSemaphore != NULL )
            {
                // See if we can obtain the semaphore.  If the semaphore is not available
                // wait 10 ticks to see if it becomes free.	
                if( xSemaphoreTake( xRetrogradeSemaphore, ( portTickType ) 10 ) == pdTRUE )

                {
                    // We were able to obtain the semaphore and can now access the
                    // shared resource.

                    if( xQueuePeek( xI2CRTCQueue, &xValues, portMAX_DELAY) == pdPASS )
                    {

                        // convert to a range of 800uS to 2200uS over 24 hours.
                        servoHours_us = (unsigned int)( ((float)xValues.Minutes + (float)xValues.Hours*60 )/1439*(2200-800) + 800);
                        // convert to a range of 800uS to 2200uS over 60 minutes.
                        servoMinutes_us = (unsigned int)( 2200 - (float)xValues.Minutes/59*(2200-800));
                  
                        set_PWM_hardware( servoHours_us, servoMinutes_us );

                        vTaskDelay( 800 / portTICK_RATE_MS ); 
    
                    }
                }   
            }
        }
        taskYIELD();
    }
}

/*-----------------------------------------------------------*/




// Convert normal decimal number byte to binary coded decimal byte
unsigned char decToBcd(unsigned char val)
{
    return ( (val/10*16) + (val%10) );
}

// Convert binary coded decimal byte to normal decimal number byte
unsigned char bcdToDec(unsigned char val)
{
    return ( (val/16*10) + (val%16) );
}



/*-----------------------------------------------------------*/

/*
    This function is only run when the time needs to be set.
    Enter the correct time (for some time in the near future).
    Compile and load the program, check the time is loading correctly.
    Then, press the restart button 4 seconds before time entered.
    Comment out the call in main() once the time is set correctly.
*/

// set the date & time to the values given here.
void setDateDs1307()
{
    unsigned char Seconds = 0;      // 0-59
    unsigned char Minutes = 00;     // 0-59
    unsigned char Hours = 22;       // 1-23
    unsigned char Day = 5;          // Sun=1, Mon=2, Tue=3, Wed=4, Thur=5, Fri=6, Sat=7
    unsigned char Date = 01;        // 1-28/29/30/31
    unsigned char Month = 7;        // Jan=1,... Dec=12
    unsigned char Year = 10;        // '00 - '99
    unsigned char ret;              // temporary  

    ret = i2c_start(DS1307+I2C_WRITE);       // set device address and write mode              
                      
    if ( ret )
    {
        /* failed to issue start condition, possibly no device found */
        i2c_stop();
        set_digital_output( IO_D3, 1);               // red LED on to show error */
        
    } else {
        /* issuing start condition ok, device accessible */

        i2c_write(0x00);                            // write address = 0 (Seconds)

        i2c_write( decToBcd( Seconds ) );  // write one byte to address 0, 0 to bit 7 starts the clock
        i2c_write( decToBcd( Minutes ) );         // write one byte to address 1
        i2c_write( decToBcd( Hours   ) );   // write one byte to address 2
        i2c_write( decToBcd( Day     ) );         // write one byte to address 3
        i2c_write( decToBcd( Date    ) );         // write one byte to address 4
        i2c_write( decToBcd( Month   ) );          // write one byte to address 5
        i2c_write( decToBcd( Year    ) );          // write one byte to address 6
        i2c_write( 0x10                );          // write one last byte to address 7,
                        // sends 0x10 (hex) 00010000 (binary) to control register - turns on square wave
        i2c_stop();                                 // set stop condition = release bus
    }
}

